/*
 * File:        Python control scio3
 *             
 * Author:      Bruce Land
 * For use with Sean Carroll's Big Board
 * http://people.ece.cornell.edu/land/courses/ece4760/PIC32/target_board.html
 * Target PIC:  PIC32MX250F128B
 * 
 * This scope takes INPUT at RB13
 * INPUT (for testing) has 10k to Vdd and 10k to sig generator
 * This gives a input range of +/-3 volts
 * For testing connect through 300 ohm resistor to signal gen
 * Vmin=0 Vmax=3.3
 */
// =============================================
// NOTE!! -- to use serial spawned functions
// you MUST EDIT config_1_3_2 to
// (1) uncomment the line -- #define use_uart_serial
// (2) SET the baud rate to match the PC terminal
// =============================================
////////////////////////////////////
// clock AND protoThreads configure!
// You MUST check this file!
#include "config_1_3_2.h"
// threading library
#include "pt_cornell_1_3_2_python.h"
#include <math.h>

////////////////////////////////////
// graphics libraries
// SPI channel 1 connections to TFT
#include "tft_master.h"
#include "tft_gfx.h"

// === the fixed point macros ========================================
typedef signed short fix14 ;
#define multfix14(a,b) ((fix14)((((long)(a))*((long)(b)))>>14)) //multiply two fixed 2.14
#define float2fix14(a) ((fix14)((a)*16384.0)) // 2^14
#define fix2float14(a) ((float)(a)/16384.0)
#define absfix14(a) abs(a) 
// === input arrays ==================================================
#define nSamp 512
#define nPixels 319
fix14 v_in[nSamp] ;

////////////////////////////////////
// === print a line on TFT =====================================================
// print string buffer
char tft_str_buffer[60];
// SEE 
// http://people.ece.cornell.edu/land/courses/ece4760/PIC32/index_TFT_display.html
// for details
void tft_printLine(int line_number, int indent, char* print_buffer, short text_color, short back_color, short char_size){
    // print_buffer is the string to print
    int v_pos, h_pos;
    char_size = (char_size>0)? char_size : 1 ;
    //
    v_pos = line_number * 8 * char_size ;
    h_pos = indent * 6 * char_size ;
    // erase the pixels
    //tft_fillRoundRect(0, v_pos, 239, 8, 1, back_color);// x,y,w,h,radius,color
    tft_setTextColor2(text_color, back_color); 
    tft_setCursor(h_pos, v_pos);
    tft_setTextSize(char_size);
    tft_writeString(print_buffer);
}

// === outputs from python handler =============================================
// signals from the python handler thread to other threads
// These will be used with the prototreads PT_YIELD_UNTIL(pt, condition);
// to act as semaphores to the processing threads
char new_string = 0;
char new_button = 0;
char new_toggle = 0;
char new_slider = 0;
char new_list = 0 ;
char new_radio = 0 ;
// identifiers and values of controls
// curent button
char button_id, button_value ;
// current toggle switch/ check box
char toggle_id, toggle_value ;
// current radio-group ID and button-member ID
char radio_group_id, radio_member_id ;
// current slider
int slider_id;
float slider_value ; // value could be large
// current listbox
int list_id, list_value ; 
// current string
char receive_string[64];

// === string input thread =====================================================
// process text from python
static PT_THREAD (protothread_python_string(struct pt *pt))
{
    PT_BEGIN(pt);
    
    // 
    while(1){
        // wait for a new string from Python
        PT_YIELD_UNTIL(pt, new_string==1);
        new_string = 0;
        
    } // END WHILE(1)   
    PT_END(pt);  
} // thread python_string

// === Buttons thread ==========================================================
// process buttons from Python 
int arm_single = 0 ;
//
static PT_THREAD (protothread_buttons(struct pt *pt))
{
    PT_BEGIN(pt);
    // set up LED port A0 to blink
    mPORTAClearBits(BIT_0 );	//Clear bits to ensure light is off.
    mPORTASetPinsDigitalOut(BIT_0);    //Set port as output
    while(1){
        PT_YIELD_UNTIL(pt, new_button==1);
        // clear flag
        new_button = 0;   
        
        // Button one -- 
        if (button_id == 1){
            arm_single = button_value ;
            //mPORTASetBits(BIT_0 );
        }
           
    } // END WHILE(1)   
    PT_END(pt);  
} // thread blink

// === Toggle thread ==========================================================
// process toggle from Python to change a dot color on the LCD
int use_trig =0 ;
int lock_ref = 0;
static PT_THREAD (protothread_toggles(struct pt *pt))
{
    PT_BEGIN(pt);
    
    //
    while(1){
        // this threaqd does a periodic redraw in case the dot is erased
        PT_YIELD_TIME_msec(100)
        //update dot color if toggle changed
        if (new_toggle == 1){
            // clear toggle flag
            new_toggle = 0;   
            if (toggle_id == 1){
                use_trig = toggle_value ;
            }
            if (toggle_id == 2){
                lock_ref = toggle_value ;
            }
           //if (toggle_id == 2){
           //     use_cursor = toggle_value ;
           // }
        } // end new toggle
        // redraw if no new event
        
    } // END WHILE(1)   
    PT_END(pt);  
} // thread toggles

// ===  Slider thread =========================================================
// process slider from Python 
int v_pos = 190 ;
int trig_level = 512 ;
short cursor_x_pos = 0 ;
// shift right 3
char v_gain = 3;
float v_pos_scale[] = {400, 200, 100, 0};

static PT_THREAD (protothread_sliders(struct pt *pt))
{
    PT_BEGIN(pt);
  
    while(1){
        PT_YIELD_UNTIL(pt, new_slider==1);
        // clear flag
        new_slider = 0; 
        if (slider_id == 1) v_pos = (int)(slider_value+v_pos_scale[v_gain]) ;
        if (slider_id == 2) trig_level = 512 + slider_value*1024/3.3 ;
        if (slider_id == 3) cursor_x_pos = (short)slider_value ;
    } // END WHILE(1)   
    PT_END(pt);  
} // thread slider

// ===  listbox thread =========================================================
// process listbox from Python 
// table gives the right-shift of the raw ADC to fit on screen
char v_gain_table[4] = {3, 2, 1, 0};
// table gives sample rates

// '0.5mS','1mS','2mS','5mS','10mS','20mS','50mS', '100mS','200mS'
int sample_freq_table[] = {320/0.0005, 320/0.001, 320/0.002, 320/0.005, 320/0.01, 320/0.02, 320/0.05, 320/0.1, 320/0.2} ;
int sample_freq = 320/0.0005 ;
//
static PT_THREAD (protothread_listbox(struct pt *pt))
{
    PT_BEGIN(pt);
    while(1){
        PT_YIELD_UNTIL(pt, new_list==1);
        // clear flag
        new_list = 0; 
        if (list_id == 1) v_gain = v_gain_table[list_value] ;
        //
        if (list_id == 2){
            sample_freq = sample_freq_table[list_value] ;
            WritePeriod3(pb_clock/sample_freq);
        }
        
    } // END WHILE(1)   
    PT_END(pt);  
} // thread listbox

// ===  radio thread =========================================================
// process listbox from Python 
int run_mode = 1;
static PT_THREAD (protothread_radio(struct pt *pt))
{
    PT_BEGIN(pt);
    while(1){
        PT_YIELD_UNTIL(pt, new_radio==1);
        // clear flag
        new_radio = 0; 
        if (radio_group_id == 1){
            
            if (radio_member_id == 1){ 
                run_mode = 1;
                mPORTASetBits(BIT_0 );
            }
            if (radio_member_id == 2){ 
                run_mode = 0;
                mPORTAClearBits(BIT_0 );
            }
        }
    } // END WHILE(1)   
    PT_END(pt);  
} // thread radio

// ===  command thread =========================================================
// process cmd TO Python 
static PT_THREAD (protothread_cmd(struct pt *pt))
{
    PT_BEGIN(pt);
    static int sys_time = 0 ;
    static char sys_led = 0 ;
    while(1){
        PT_YIELD_TIME_msec(500) ;
        // send a command
        // $ + event# 00-99 + space + value
       // printf("$01 %d\r", sys_time++) ;
        // $ + event# 00-99 + space + value
        // blink a virtual LED
        sys_led = !sys_led ;
        // the format %1d means print only one (low order) digit
        printf("$02%1d\r", sys_led) ;
        
        //PT_YIELD_TIME_msec(500) ;
        //sys_led = !sys_led ;
        //printf("$02%1d\r", sys_led) ;
        
    } // END WHILE(1)   
    PT_END(pt);  
} // thread cmd

// ===  draw thread =========================================================
// draw traces on tft
// DMA channel busy flag
#define CHN_BUSY 0x80
// use_trig 1 if trigger level is used
// trig_level voltage trigger point
// run_mode 1 if muliple trace trace
// arm_single 1 when users want a trace in single mode
static PT_THREAD (protothread_draw(struct pt *pt))
{
    PT_BEGIN(pt);
    static int pixels[nPixels] ;
    static int sample_number ;
    static int trig1, trig2 ;
    static int v_min, v_max ;
    static int print_time ;
    static int trig_time ;
    static int Tref, Vref ;
    //
    while(1){
        PT_YIELD_TIME_msec(10) ;
        //PT_YIELD(pt) ;
        if (print_time++ >20) print_time = 0 ;
        
        trig1 = 1024;
        trig2 = 0 ;  
        if (use_trig == 1){
            trig_time = PT_GET_TIME() ;
            while(!(trig2>trig_level && trig1<=trig_level)) {                
                // wait for ADC done
                 while (!IFS1 & 0x0002) {};
                // first conversion
                // Read the value
                trig1 = ADC1BUF0 ; //IFS1<1>

                // wait for next one
                 while (!IFS1 & 0x0002) {};
                // second conversion
                // Read the value
                trig2 = ADC1BUF0 ;
                // bail out
                if (PT_GET_TIME()> (trig_time + 500)) break ;
            }
        }
        // else just acquire a trace
        
        if (run_mode || arm_single){
            // enable ADC DMA channel and get nSamp samples from ADC
            DmaChnEnable(3);
            //while((DCH3CON & CHN_BUSY) ){};
            PT_YIELD_UNTIL(pt, !(DCH3CON & CHN_BUSY));
            // if armed clear it
            arm_single = 0;
        }
        // void tft_drawRect(short x, short y, short w, short h, unsigned short color);
        // Display on TFT
        // v_in[sample_number]>>2 is approx full screen 3.0v input
        // erase, then draw
        v_min = 1024; v_max = 0;
        for (sample_number = 0; sample_number < nPixels; sample_number++) {
            tft_drawPixel(sample_number, pixels[sample_number], ILI9340_BLACK);
            //tft_drawPixel(sample_number, pixels[sample_number]+1, ILI9340_BLACK);
            // negate v_in so that UP is +
            pixels[sample_number] = -(v_in[sample_number]>>v_gain) + v_pos ; //v_in[sample_number]
            tft_drawPixel(sample_number, pixels[sample_number], ILI9340_WHITE); 
            //tft_drawPixel(sample_number, pixels[sample_number]+1, ILI9340_WHITE); 
            
            if (v_in[sample_number] < v_min) v_min = v_in[sample_number];
            if (v_in[sample_number] > v_max) v_max = v_in[sample_number];
        } 
        
        PT_YIELD(pt);
        int st, ed, mid_ADC ;
        st=0; ed=0; 
        // compute middle of voltage range
        mid_ADC = v_min + ((v_max-v_min)>>1);
        // find where waveform passes thru mid_ADC
        for (sample_number=0; sample_number<nSamp; sample_number++){
            if (v_in[sample_number]<mid_ADC && v_in[sample_number+1]>=mid_ADC && st==0 ) st = sample_number;
            if (v_in[sample_number]<mid_ADC && v_in[sample_number+1]>=mid_ADC && st>0 && sample_number>st+2 && ed==0) ed = sample_number;
        }

        int f ;
        //  print the frequency 
        if (print_time == 1 && lock_ref == 0){          
            if (ed-st > 10) f = sample_freq/(ed-st);
            else f = 0 ;
            // send a command
            // $ + event# 00-99 + space + value
            printf("$03 %d\r", f) ;
        }
        PT_YIELD(pt);
        // run the cursor 1
        // tft_drawFastVLine(short x, short y, short h, unsigned short color)
        
        static int old_local_cursor_x = 0 ;
        static int old_local_vpos = 0 ;
        tft_drawFastVLine(old_local_cursor_x, 5, 230, ILI9340_BLACK);
        tft_drawFastHLine(old_local_cursor_x,  old_local_vpos , 20, ILI9340_BLACK);
        old_local_cursor_x = cursor_x_pos ;
        old_local_vpos = v_pos -(512>>v_gain) ;
        tft_drawFastVLine(cursor_x_pos, 5, 230, ILI9340_CYAN);
        tft_drawFastHLine(cursor_x_pos,  v_pos-(512>>v_gain) , 20, ILI9340_CYAN);
        //
        if (lock_ref == 0){
            if (print_time == 2){
                 printf("$04 %d\r", v_in[cursor_x_pos]-512) ;
                 Vref = v_in[cursor_x_pos]-512;
            }
            PT_YIELD(pt);
            // the time
            if (print_time == 3){
                printf("$01 %7.3f\r", 1000. * (float)cursor_x_pos / (float)sample_freq ) ;
                Tref = cursor_x_pos ;
            }
        }
        if (lock_ref == 1){
            if (print_time == 2){
                 printf("$04 %d\r", v_in[cursor_x_pos]-512-Vref) ;
            }
            PT_YIELD(pt);
            // the time
            if (print_time == 3){
                printf("$01 %7.3f\r", 1000. * (float)(cursor_x_pos-Tref) / (float)sample_freq ) ;
            }
            if (print_time == 1){             
                if (abs(cursor_x_pos-Tref) > 5) f = abs(sample_freq/(cursor_x_pos-Tref));
                else f = 0 ;
                printf("$03 %d\r", f) ;
            }
        }
    } // END WHILE(1)   
    PT_END(pt);  
} // thread draw

// === Python serial thread ====================================================
// you should not need to change this thread UNLESS you add new control types
static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
    static char junk;
    //   
    //
    while(1){
        // There is no YIELD in this loop because there are
        // YIELDS in the spawned threads that determine the 
        // execution rate while WAITING for machine input
        // =============================================
        // NOTE!! -- to use serial spawned functions
        // you MUST edit config_1_3_2 to
        // (1) uncomment the line -- #define use_uart_serial
        // (2) SET the baud rate to match the PC terminal
        // =============================================
        
        // now wait for machine input from python
        // Terminate on the usual <enter key>
        PT_terminate_char = '\r' ; 
        PT_terminate_count = 0 ; 
        PT_terminate_time = 0 ;
        // note that there will NO visual feedback using the following function
        PT_SPAWN(pt, &pt_input, PT_GetMachineBuffer(&pt_input) );
        
        // Parse the string from Python
        // There can be toggle switch, button, slider, and string events
        
        // toggle switch
        if (PT_term_buffer[0]=='t'){
            // signal the button thread
            new_toggle = 1;
            // subtracting '0' converts ascii to binary for 1 character
            toggle_id = (PT_term_buffer[1] - '0')*10 + (PT_term_buffer[2] - '0');
            toggle_value = PT_term_buffer[3] - '0';
        }
        
        // pushbutton
        if (PT_term_buffer[0]=='b'){
            // signal the button thread
            new_button = 1;
            // subtracting '0' converts ascii to binary for 1 character
            button_id = (PT_term_buffer[1] - '0')*10 + (PT_term_buffer[2] - '0');
            button_value = PT_term_buffer[3] - '0';
        }
        
        // slider
        if (PT_term_buffer[0]=='s'){
            sscanf(PT_term_buffer, "%c %d %f", &junk, &slider_id, &slider_value);
            new_slider = 1;
        }
        
        // listbox
        if (PT_term_buffer[0]=='l'){
            new_list = 1;
            list_id = PT_term_buffer[2] - '0' ;
            list_value = PT_term_buffer[3] - '0';
            //printf("%d %d", list_id, list_value);
        }
        
        // radio group
        if (PT_term_buffer[0]=='r'){
            new_radio = 1;
            radio_group_id = PT_term_buffer[2] - '0' ;
            radio_member_id = PT_term_buffer[3] - '0';
            //printf("%d %d", radio_group_id, radio_member_id);
        }
        
        // string from python input line
        if (PT_term_buffer[0]=='$'){
            // signal parsing thread
            new_string = 1;
            // output to thread which parses the string
            // while striping off the '$'
            strcpy(receive_string, PT_term_buffer+1);
        }                                  
    } // END WHILE(1)   
    PT_END(pt);  
} // thread blink

// === Main  ======================================================

void main(void) {
   
   // === init FFT data =====================================
    // one cycle sine table
    //  required for FFT
    //int ii;
    //for (ii = 0; ii < N_WAVE; ii++) {
    //    Sinewave[ii] = float2fix14(sin(6.283 * ((float) ii) / N_WAVE)*0.5);
     ///   window[ii] = float2fix14(1.0 * (1.0 - cos(6.283 * ((float) ii) / (N_WAVE - 1))));
        //window[ii] = float2fix(1.0) ;
    //}
 
  
  // the ADC ///////////////////////////////////////
    
    // timer 3 setup for adc trigger  ==============================================
    // Set up timer3 on,  no interrupts, internal clock, prescalar 1, compare-value
    // This timer generates the time base for each ADC sample. 
    // works at ??? Hz
    
    // PB clock rate
    OpenTimer3(T3_ON | T3_SOURCE_INT | T3_PS_1_1, pb_clock/sample_freq); 
    
    //=== DMA Channel 0 transfer ADC data to array v_in ================================
    // Open DMA Chan1 for later use sending video to TV
    #define DMAchan3 3
	DmaChnOpen(DMAchan3, 0, DMA_OPEN_DEFAULT);
    DmaChnSetTxfer(DMAchan3, (void*)&ADC1BUF0, (void*)v_in, 2, nSamp*2, 2); //512 16-bit integers
    DmaChnSetEventControl(DMAchan3, DMA_EV_START_IRQ(28)); // 28 is ADC done
    // ==============================================================================
    
    // configure and enable the ADC
    CloseADC10(); // ensure the ADC is off before setting the configuration

    // define setup parameters for OpenADC10
    // Turn module on | ouput in integer | trigger mode auto | enable autosample
    // ADC_CLK_AUTO -- Internal counter ends sampling and starts conversion (Auto convert)
    // ADC_AUTO_SAMPLING_ON -- Sampling begins immediately after last conversion completes; SAMP bit is automatically set
    // ADC_AUTO_SAMPLING_OFF -- Sampling begins with AcquireADC10();
    #define PARAM1  ADC_FORMAT_INTG16 | ADC_CLK_TMR | ADC_AUTO_SAMPLING_ON //

    // define setup parameters for OpenADC10
    // ADC ref external  | disable offset test | disable scan mode | do 1 sample | use single buf | alternate mode off
    #define PARAM2  ADC_VREF_AVDD_AVSS | ADC_OFFSET_CAL_DISABLE | ADC_SCAN_OFF | ADC_SAMPLES_PER_INT_1 | ADC_ALT_BUF_OFF | ADC_ALT_INPUT_OFF
    //
    // Define setup parameters for OpenADC10
    // for a 40 MHz PB clock rate
    // use peripherial bus clock | set sample time | set ADC clock divider
    // ADC_CONV_CLK_Tcy should work at 40 MHz.
    // ADC_SAMPLE_TIME_6 seems to work with a source resistance < 1kohm
    //#define PARAM3 ADC_CONV_CLK_PB | ADC_SAMPLE_TIME_6 | ADC_CONV_CLK_Tcy //ADC_SAMPLE_TIME_5| ADC_CONV_CLK_Tcy2
    // 30 MHz PB bus
    #define PARAM3 ADC_CONV_CLK_PB | ADC_SAMPLE_TIME_5 | ADC_CONV_CLK_Tcy2 //ADC_SAMPLE_TIME_15| ADC_CONV_CLK_Tcy2
    // define setup parameters for OpenADC10
    // set AN11 and  as analog inputs
    #define PARAM4	ENABLE_AN11_ANA // RB13

    // define setup parameters for OpenADC10
    // do not assign channels to scan
    #define PARAM5	SKIP_SCAN_ALL

    // use ground as neg ref for A | use AN11 for input A     
    // configure to sample AN11 
    SetChanADC10(ADC_CH0_NEG_SAMPLEA_NVREF | ADC_CH0_POS_SAMPLEA_AN11); // configure to sample AN4 
    OpenADC10(PARAM1, PARAM2, PARAM3, PARAM4, PARAM5); // configure ADC using the parameters defined above

    EnableADC10(); // Enable the ADC
    ///////////////////////////////////////////////////////
    
  // === setup system wide interrupts  ========
  INTEnableSystemMultiVectoredInt();
  
  // === TFT setup ============================
  // init the display in main since more than one thread uses it.
  // NOTE that this init assumes SPI channel 1 connections
  tft_init_hw();
  tft_begin();
  tft_fillScreen(ILI9340_BLACK);
  //240x320 vertical display
  tft_setRotation(1); // Use tft_setRotation(1) for 320x240
  
  // === config threads ========================
  PT_setup();
  
  // === identify the threads to the scheduler =====
  // add the thread function pointers to be scheduled
  // --- Two parameters: function_name and rate. ---
  // rate=0 fastest, rate=1 half, rate=2 quarter, rate=3 eighth, rate=4 sixteenth,
  // rate=5 or greater DISABLE thread!
  
  pt_add(protothread_buttons, 0);
  pt_add(protothread_serial, 0);
  pt_add(protothread_python_string, 0);
  pt_add(protothread_toggles, 0);
  pt_add(protothread_sliders, 0);
  pt_add(protothread_listbox, 0);
  pt_add(protothread_radio, 0);
  pt_add(protothread_cmd, 0);
  // TFT drawing thread
  pt_add(protothread_draw, 0);
  
  // === initalize the scheduler ====================
  PT_INIT(&pt_sched) ;
  // >>> CHOOSE the scheduler method: <<<
  // (1)
  // SCHED_ROUND_ROBIN just cycles thru all defined threads
  //pt_sched_method = SCHED_ROUND_ROBIN ;
  
  // NOTE the controller must run in SCHED_ROUND_ROBIN mode
  // ALSO note that the scheduler is modified to cpy a char
  // from uart1 to uart2 for the controller
  
  pt_sched_method = SCHED_ROUND_ROBIN ;
  
  // === scheduler thread =======================
  // scheduler never exits
  PT_SCHEDULE(protothread_sched(&pt_sched));
  // ============================================
  
} // main
// === end  ======================================================

